
@file:Suppress("unused")

package ru.dbotthepony.mc.otm.core.math

import com.mojang.math.Matrix3f
import com.mojang.math.Matrix4f
import com.mojang.math.Vector3f
import com.mojang.math.Vector4f

fun Matrix4f.rotate(angle: IAngle): Matrix4f {
	multiply(angle.rotationXYZW())
	return this
}

fun Matrix4f.translate(vector: Vector) {
	translate(-vector.asVector3f())
}

fun Matrix4f.rotateAroundPoint(point: Vector, axis: Vector, rotation: Float, isDegrees: Boolean = false) {
	val p = point.asVector3f()
	translate(-p)
	multiply(axis.rotateAroundThis(rotation, isDegrees))
	translate(-p)
}

fun Matrix4f.rotateAroundPoint(point: Vector, axis: Vector3f, rotation: Float, isDegrees: Boolean = false) {
	val p = point.asVector3f()
	translate(-p)
	multiply(if (isDegrees) axis.rotationDegrees(rotation) else axis.rotation(rotation))
	translate(-p)
}

fun Matrix4f.rotateAroundPoint(point: Vector, rotation: IAngle) {
	val p = point.asVector3f()
	translate(-p)
	multiply(rotation.rotationXYZW())
	translate(-p)
}

fun Matrix4f.rotateAroundPoint(point: Vector3f, axis: Vector, rotation: Float, isDegrees: Boolean = false) {
	translate(-point)
	multiply(axis.rotateAroundThis(rotation, isDegrees))
	translate(-point)
}

fun Matrix4f.rotateAroundPoint(point: Vector3f, rotation: IAngle) {
	translate(-point)
	multiply(rotation.rotationXYZW())
	translate(-point)
}

fun Matrix4f.rotateAroundPoint(point: Vector3f, axis: Vector3f, rotation: Float, isDegrees: Boolean = false) {
	translate(-point)
	multiply(if (isDegrees) axis.rotationDegrees(rotation) else axis.rotation(rotation))
	translate(-point)
}

// 1.19.3 <-> 1.19.2 behavior
// mojang decided it is a good idea to make their matrices row-major
// when basically everyone else (well, aside from thing nobody uses, what is it called, uuuuuuuugh directx i think)
// uses column-major matrices
operator fun Matrix4f.get(column: Int, row: Int): Float {
	require(row in 0 .. 3) { "Invalid row: $row" }
	require(column in 0 .. 3) { "Invalid column: $column" }

	return when (val index = row or (column shl 2)) {
		0 or (0 shl 2) -> m00
		1 or (0 shl 2) -> m10
		2 or (0 shl 2) -> m20
		3 or (0 shl 2) -> m30
		0 or (1 shl 2) -> m01
		1 or (1 shl 2) -> m11
		2 or (1 shl 2) -> m21
		3 or (1 shl 2) -> m31
		0 or (2 shl 2) -> m02
		1 or (2 shl 2) -> m12
		2 or (2 shl 2) -> m22
		3 or (2 shl 2) -> m32
		0 or (3 shl 2) -> m03
		1 or (3 shl 2) -> m13
		2 or (3 shl 2) -> m23
		3 or (3 shl 2) -> m33
		else -> throw IllegalStateException("$index")
	}
}

operator fun Matrix4f.set(column: Int, row: Int, value: Float) {
	require(row in 0 .. 3) { "Invalid row: $row" }
	require(column in 0 .. 3) { "Invalid column: $column" }

	when (val index = row or (column shl 2)) {
		0 or (0 shl 2) -> m00 = value
		1 or (0 shl 2) -> m10 = value
		2 or (0 shl 2) -> m20 = value
		3 or (0 shl 2) -> m30 = value
		0 or (1 shl 2) -> m01 = value
		1 or (1 shl 2) -> m11 = value
		2 or (1 shl 2) -> m21 = value
		3 or (1 shl 2) -> m31 = value
		0 or (2 shl 2) -> m02 = value
		1 or (2 shl 2) -> m12 = value
		2 or (2 shl 2) -> m22 = value
		3 or (2 shl 2) -> m32 = value
		0 or (3 shl 2) -> m03 = value
		1 or (3 shl 2) -> m13 = value
		2 or (3 shl 2) -> m23 = value
		3 or (3 shl 2) -> m33 = value
		else -> throw IllegalStateException("$index")
	}
}

operator fun Matrix3f.get(column: Int, row: Int): Float {
	require(row in 0 .. 2) { "Invalid row: $row" }
	require(column in 0 .. 2) { "Invalid column: $column" }

	return when (val index = row or (column shl 2)) {
		0 or (0 shl 2) -> m00
		1 or (0 shl 2) -> m10
		2 or (0 shl 2) -> m20
		0 or (1 shl 2) -> m01
		1 or (1 shl 2) -> m11
		2 or (1 shl 2) -> m21
		0 or (2 shl 2) -> m02
		1 or (2 shl 2) -> m12
		2 or (2 shl 2) -> m22
		else -> throw IllegalStateException("$index")
	}
}

fun Matrix3f.mset(column: Int, row: Int, value: Float) {
	require(row in 0 .. 2) { "Invalid row: $row" }
	require(column in 0 .. 2) { "Invalid column: $column" }

	when (val index = row or (column shl 2)) {
		0 or (0 shl 2) -> m00 = value
		1 or (0 shl 2) -> m10 = value
		2 or (0 shl 2) -> m20 = value
		0 or (1 shl 2) -> m01 = value
		1 or (1 shl 2) -> m11 = value
		2 or (1 shl 2) -> m21 = value
		0 or (2 shl 2) -> m02 = value
		1 or (2 shl 2) -> m12 = value
		2 or (2 shl 2) -> m22 = value
		else -> throw IllegalStateException("$index")
	}
}

fun Matrix4f.identity() = setIdentity()
fun Matrix3f.identity() = setIdentity()

fun Matrix4f.identityFast() {
	m00 = 1.0F
	m11 = 1.0F
	m22 = 1.0F
	m33 = 1.0F
}

fun Matrix3f.identityFast() {
	m00 = 1.0F
	m11 = 1.0F
	m22 = 1.0F
}

fun Matrix3f.toMatrix4f(): Matrix4f {
	val result = Matrix4f()

	result.m00 = this.m00
	result.m10 = this.m10
	result.m20 = this.m20
	result.m01 = this.m01
	result.m11 = this.m11
	result.m21 = this.m21
	result.m02 = this.m02
	result.m12 = this.m12
	result.m22 = this.m22

	return result
}

var Matrix4f.translation: Vector3f
	get() {
		return Vector3f(this[0, 3], this[1, 3], this[2, 3])
	}

	set(value) {
		this[0, 0] = 1f
		this[1, 1] = 1f
		this[2, 2] = 1f
		this[3, 3] = 1f

		this[0, 3] = value.x()
		this[1, 3] = value.y()
		this[2, 3] = value.z()
	}

var Matrix4f.translation4: Vector4f
	get() {
		return Vector4f(this[1, 3], this[1, 3], this[2, 3], this[3, 3])
	}

	set(value) {
		this[0, 0] = 1f
		this[1, 1] = 1f
		this[2, 2] = 1f
		this[3, 3] = 1f

		this[0, 3] = value.x()
		this[1, 3] = value.y()
		this[2, 3] = value.z()
		this[3, 3] = value.w()
	}
